# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_type.ipynb.

# %% auto 0
__all__ = ['T', 'O', 'P', 'Ts']

# %% ../nbs/01_type.ipynb 6
#| export


# %% ../nbs/01_type.ipynb 8
from typing import (TypeVar, ParamSpec, TypeVarTuple)

# %% ../nbs/01_type.ipynb 10
#| export

# %% ../nbs/01_type.ipynb 12
#| export

# %% ../nbs/01_type.ipynb 14
#| export

# %% ../nbs/01_type.ipynb 16
T = TypeVar('T')
'''Type variable for generic types''';

O = TypeVar('O')
'''Type variable for generic types reprenesting "other"''';

P = ParamSpec('P')
'''Parameter specification type variable, capturing the type of function parameters.''';

Ts = TypeVarTuple('Ts')
